#######################################
## Bivariate Polynomial Gaussian CWM ##
#######################################

source("ML.R")

PolyCWM <- function(
	Y,			                        # Y: a numerical vector for the response variable
  X,			                        # X: a numerical vector for the unique covariate
	k,                              # number of groups
	r,                              # polinomial order
	zlab = NULL, 		                # TRUE classification for the labeled data (which are the first nlab observations in X and Y)
	algorithm = "EM",               # algorithm used (options: "EM" or "CEM")
	method = "BFGS",                # optimization method in each EM iteration
  initialization = "random.soft", # initialization procedure: "random.soft", "random.hard", and "manual" 
	startgroups = NULL,             # vector of initial group-labels: it is used only if initialization="manual"
	#start.z = NULL,                # (n x k)-matrix of soft or hard classification: it is used only if initialization="manual"		
	iter.max = 1000,                # maximum number of iterations in the EM-algorithm
	threshold = 1.0e-04,            # stopping rule in the Aitken rule
	loglikplot = TRUE,              # if TRUE, the log-likelihood values against the iterations are plotted
	SE = FALSE                      # if SE = TRUE, then Standard Errors are computed
	)
{

  library(mclust)
  
  n      <- length(Y)       # sample size
  nlab   <- length(zlab)/k  # number of labelled observations
  nunlab <- n - nlab        # number of unlabelled observations
  
  # weights
  
  prior  <- numeric(k) # weights
    
  # for P(X|j)
  
  muX <- array(0,c(k),dimnames=list(paste("comp.",1:k,sep="")))        # mean of X in each group
  sdX <- array(0,c(k),dimnames=list(paste("comp.",1:k,sep="")))        # sd of X in each group
  PX  <- array(0,c(n,k),dimnames=list(1:n,paste("comp.",1:k,sep=""))) 
  
  # for P(Y|X,j)
  
  beta <- array(0,c(k,r+1),dimnames=list(paste("comp.",1:k,sep=""),paste("beta.",0:r,sep=""))) # (k x (r+1)) matrix of coefficients (plus intercept) for each group
  sdY  <- array(0,c(k),dimnames=list(paste("comp.",1:k,sep="")))                               # conditional standard deviation in each group
  PY   <- array(0,c(n,k),dimnames=list(1:n,paste("comp.",1:k,sep="")))
  muY  <- array(0,c(n,k),dimnames=list(1:n,paste("comp.",1:k,sep="")))                         # conditional mean in each group (for each covariate value)
  
  # ------------------------- #
  # posteriors initialization #
  # ------------------------- #
  
  # RANDOM INITIALIZATION
  
  if(initialization=="random.soft"){
    
    post <- array(runif(nunlab*k),c(nunlab,k))  
    post <- post/rowSums(post)                 
    z    <- rbind(zlab,post)                   # probabilities of group membership
    if(algorithm=="CEM") 
      z <- class(map(z),k=k)
            
  } 
  
  if(initialization=="random.hard"){
    
    zunlab <- t(rmultinom(nunlab, size = 1, prob=rep(1/k,k)))  
    z      <- rbind(zlab,zunlab)                              # group membership
        
  } 
  
  if(initialization=="manual"){ 
    
    z <- class(startgroups,k=k)
    if(nlab>0)
      z[1:nlab,] <- zlab        # group membership
    
  }
  
  # ------------ #
  # EM algorithm #
  # ------------ #
  
  # Preliminary definition of convergence criterions
  
  check     <- 0
  iteration <- 1
  loglik    <- NULL
  aloglik   <- NULL
  aloglik   <- c(0,0)
  a         <- NULL
  a         <- c(0,0)
  
  while(check<1){
    
    # ++++++ #
    # M-step #
    # ++++++ #
    
    # ------- #
    # Weights #
    # ------- #
    
    prior <- colMeans(z)
    
    # - #
    # X #
    # - #
    
    #print("X")
    
    for(j in 1:k){
      modelX  <- GaussianX(X,weights=z[,j])
      muX[j]  <- modelX$muX
      sdX[j]  <- modelX$sdX
      PX[,j]  <- dnorm(X,mean=muX[j],sd=sdX[j])              
    }
    PX  <- (PX<10^(-322))*10^(-322)+(PX>10^(-322))*PX # to avoid numerical problems 
        
    # --- #
    # Y|x #
    # --- #
    
    #print("Y|x")
    
    for(j in 1:k){
      modelY      <- GaussianY(Y,X,r=r,weights=z[,j])
      beta[j,]    <- modelY$beta
      sdY[j]      <- modelY$sdY
      muY[,j]     <- polyfun(X,par=beta[j,])
      PY[,j]      <- dnorm(Y,mean=muY[,j],sd=sdY[j])              
    }
    PY  <- (PY<10^(-322))*10^(-322)+(PY>10^(-322))*PY # to avoid numerical problems 
        
    # -------------- # 
    # log-likelihood # 
    # -------------- #
    
    if(algorithm=="EM" && identical(nlab,0)){
      
      llvalues    <- sum(log(rowSums(matrix(rep(prior,n),n,k,byrow=TRUE)*PY*PX))) 
      
    }
    
    if(algorithm=="EM" && nlab!=0){
      
      PXYjlab     <- zlab*(log(matrix(rep(prior,nlab),nlab,k,byrow=T))+log(PY[1:nlab,])+log(PX[1:nlab,]))
      PXYlab      <- rowSums(PXYjlab)
      PXYjunlab   <- matrix(rep(prior,nunlab),nunlab,k,byrow=T)*PY[(nlab+1):n,]*PX[(nlab+1):n,]
      PXYunlab    <- rowSums(PXYjunlab)
      llvalues    <- sum(PXYlab)+sum(log(PXYunlab))
      
    }
    
    if(algorithm=="CEM"){
      
      PXYj        <- z*(log(matrix(rep(prior,n),n,k,byrow=T))+log(PY)+log(PX))
      PXY         <- rowSums(PXYj)
      llvalues    <- sum(PXY)
    }
    
    loglik[iteration] <- llvalues
    
    # ---------------------------------------------- #
    # Aitken's Acceleration-Based Stopping Criterion #
    # ---------------------------------------------- #
    
    if(iteration>2){
      a[iteration-1]      <- (loglik[iteration]-loglik[iteration-1])/(loglik[iteration-1]-loglik[iteration-2])
      aloglik[iteration]  <- loglik[iteration-1]+(1/(1-a[iteration-1])*(loglik[iteration]-loglik[iteration-1]))
      if(abs(aloglik[iteration]-loglik[iteration])<threshold) 
        check <- 1 
    }
    
    if(iteration==iter.max | k==1) check <- 1
    
    cat("Iteration",iteration,"\n")
    iteration <- iteration + 1
    
    # ++++++ #
    # E-Step #
    # ++++++ #
    
    z.num  <- matrix(rep(prior,n),n,k,byrow=TRUE)*PY*PX  # (n x k)
    z.den  <- rowSums(z.num)                             # n-vector
    z      <- z.num/matrix(rep(z.den,k),ncol=k)          # (n x k)
    
    if(algorithm=="CEM") 
      z <- class(map(z),k=k)
    if(nlab>0) 
      z[1:nlab,] <- zlab
    
    }
  
  finalloglik <- loglik[iteration-1] 
  
  # ----------------------------------------------------------------------- #
  # The EM-algorithm is finished                                            #
  # Check on the EM-monotonicity                                            #
  # Plotting the values of the observed loglikelihood versus the iterations #
  # ----------------------------------------------------------------------- #
  
  if(loglikplot){
    
    par(mai=c(0.84,0.8,0.012,0.004))
    par(las = 3)
    par(cex.axis=0.7)
    par(cex.lab=1.2)
    plot(0:(iteration-2),loglik[1:(iteration-1)],type="l",axes = FALSE,xlab="iterations",ylab="log-likelihood",lwd=2)
    axis(1, at = 0:(iteration-2),label = 0:(iteration-2)) 
    axis(2)
    box(col = "black")
    
  }
  
  # --------------------- #
  # Classification vector #
  # --------------------- #

  group <- apply(z,1,which.max) 
  
  # --------------------- #
  # Information criteria  #
  # --------------------- #
  
  npar      <- (k-1) + k*(r+2) + 2*k   
  
  AIC       <- 2*finalloglik - npar*2
  BIC       <- 2*finalloglik - npar*log(n)
  
  z.const   <- (z<10^(-322))*10^(-322)+(z>10^(-322))*z   # vincolo per evitare i NaN nel calcolo di tau*log(tau)
  hard.z    <- (matrix(rep(apply(z,1,max),k),n,k,byrow=F)==z)*1
  ECM       <- sum(hard.z[(nlab+1):n,]*log(z.const[(nlab+1):n,]))
  ICL       <- BIC+ECM
  
  # --------------- #
  # Standard Errors #
  # --------------- #
  
  SEprior       <- NULL
  SEmuX      <- NULL
  SEsdX      <- NULL
  SEbeta     <- NULL
  SEsdY      <- NULL
  
  if(SE){
    
    cat("\n Standard Errors Computation... please wait\n")
    
    if(algorithm=="EM"){
      
      fmin <- function(s,Y=Y,X=X,n=n,k=k,r=r,tau=z){
        
        s[1:k]                                           -> weights
        s[(k+1):(2*k)]                                   -> muX
        s[(2*k+1):(3*k)]                                 -> sdX
        matrix(s[(3*k+1):(3*k+(r+1)*k)],k,(r+1),byrow=T) -> betaY 
        s[(3*k+(r+1)*k+1):(3*k+(r+1)*k+k)]               -> sdY 
        
        # Observed-data log-likelihood
        
        if(identical(nlab,0)){
          
          l  <- 0    # observed-data log-likelihood
          for(i in 1:n){
            lj <- 0
            for(j in 1:k){
              lj <- lj + weights[j]*dnorm(X[i],mean=muX[j],sd=sdX[j],log=F)*dnorm(Y[i],mean=outer(X[i],0:r,"^")%*%betaY[j,],sd=sdY[j],log=F)
            }
            l <- l + log(lj)
          }
          
        }
        
        if(nlab>0){
          
          lx <- 0    # log-likelihood of X
          ly <- 0    # log-likelihood of Y|X
          lw <- 0    # log-likelihood of the weights
          
          for(i in 1:nlab){
            for(j in 1:k){
              lw <- lw + tau[i,j]*log(weights[j])
              lx <- lx + tau[i,j]*dnorm(X[i],mean=muX[j],sd=sdX[j],log=T)
              ly <- ly + tau[i,j]*dnorm(Y[i],mean=outer(X[i],0:r,"^")%*%betaY[j,],sd=sdY[j],log=T) 
            }
          }
          l <- lx + ly + lw  # complete-data log-likelihood
          
          for(i in 1:nunlab){
            lj <- 0
            for(j in 1:k){
              lj <- lj + weights[j]*dnorm(X[i+nlab],mean=muX[j],sd=sdX[j],log=F)*dnorm(Y[i+nlab],mean=outer(X[i+nlab],0:r,"^")%*%betaY[j,],sd=sdY[j],log=F)
            }
            l <- l + log(lj)
          }
        }
        
        return(-l)
        
      }
    }
    
    if(algorithm=="CEM"){
      
      fmin <- function(s,Y=Y,X=X,n=n,k=k,r=r,tau=post){
        
        s[1:k]                                           -> weights
        s[(k+1):(2*k)]                                   -> muX
        s[(2*k+1):(3*k)]                                 -> sdX
        matrix(s[(3*k+1):(3*k+(r+1)*k)],k,(r+1),byrow=T) -> betaY 
        s[(3*k+(r+1)*k+1):(3*k+(r+1)*k+k)]               -> sdY 
        
        # Complete-data log-likelihood
        
        lx <- 0    # log-likelihood of X
        ly <- 0    # log-likelihood of Y|X
        lw <- 0    # log-likelihood of the weights
        
        for(i in 1:n){
          for(j in 1:k){
            lw <- lw + tau[i,j]*log(weights[j])
            lx <- lx + tau[i,j]*dnorm(X[i],mean=muX[j],sd=sdX[j],log=T)
            ly <- ly + tau[i,j]*dnorm(Y[i],mean=outer(X[i],0:r,"^")%*%betaY[j,],sd=sdY[j],log=T) 
          }
        }
        l <- lx + ly + lw  # complete-data log-likelihood
        
        return(-l)
        
      }
    }
    
    start.par <- c(prior,muX,sdX,t(beta),sdY)
    lower     <- numeric(npar+1)
    upper     <- numeric(npar+1)
    tol       <- 0.00000001
    lower     <- start.par-tol
    upper     <- start.par+tol
    SE        <- optim(par=start.par,fn=fmin,hessian=TRUE,method="L-BFGS-B",lower=lower,upper=upper,Y=Y,X=X,n=n,k=k,r=r,tau=z,control=list(maxit=100)) # fnscale=-1,
    OI        <- solve(SE$hessian)
    se        <- sqrt(diag(OI))
    SEprior   <- se[1:k]
    SEmuX     <- se[(k+1):(2*k)]
    SEsdX     <- se[(2*k+1):(3*k)]
    SEbeta    <- matrix(se[(3*k+1):(3*k+(r+1)*k)],k,byrow=T)
    dimnames(SEbeta)[[1]] <- paste("comp",1:k,sep="")
    dimnames(SEbeta)[[2]] <- paste("beta",0:r,sep="")
    SEsdY     <- se[(3*k+(r+1)*k+1):(3*k+(r+1)*k+k)]
    
  }

result <- list(
  Y         = Y,            # Y: a numerical vector for the response variable
  X         = X,            # X: a numerical vector for the unique covariate
  k         = k,            # number of groups
  r         = r,            # polynomial order
  n         = n,            # sample size
  npar      = npar,
  algorithm = algorithm,    # algorithm used in ML estimation 
  prior     = prior,
  muX       = muX,
  sdX       = sdX,
  PX        = PX,
  beta      = beta,
  muY       = muY,
  sdY       = sdY,
  PY        = PY,
  iter.stop = iteration,
  z         = z,
  group     = group,
  loglik    = finalloglik,
  AIC       = AIC,
  BIC       = BIC,
  ICL       = ICL,      # alla McNicholas
  SEprior   = SEprior,
  SEmuX     = SEmuX,
  SEsdX     = SEsdX,
  SEbeta    = SEbeta,
  SEsdY     = SEsdY,
  call      = match.call()
)

class(result) <- "PolyCWM"
return(result)

alarm()

}

#################
## PLOT METHOD ##
#################

plot.PolyCWM <- function(x,plottype="CW",theta=30,nbin=35,contour=TRUE,nlevels=21,...)
{
  # theta is used only for plottype="density" and governs the orientation of the 3D-plot
  # nbin is used only for plottype="CW" and governs the number od equidistant bins
  # contour is not used for plottype="density"; if contour=TRUE, the contour lines are superimposed on the plot
  # nlevels is not used for plottype="density"; it governs the number of contour lines
  
  if(!any(plottype==c("density","scatter","CW"))){
    stop("error: the value specified for plottype is not density, scatter, or CW")}
  eval(parse(text=paste0(".",plottype,".plot(x=x,theta=theta,nbin=nbin,contour=contour,nlevels=nlevels,...)")))
  #if(CI & any(plottype==c("obsfit","fitted")& !is.null(x$ex))){.ci.plot(x=x,...)}
}

.density.plot <- function(x,theta=30,...)
{
  X       <- x$X
  Y       <- x$Y
  n       <- x$n
  k       <- x$k
  r       <- x$r
  group   <- x$group
  prior   <- x$prior
  beta    <- x$beta
  muX     <- x$muX
  sdX     <- x$sdX
  beta    <- x$beta
  sdY     <- x$sdY
   
  prec    <- 101
  perc    <- 0.1
  gridx <- seq((min(X)-perc*(max(X)-min(X))),(max(X)+perc*(max(X)-min(X))),length=prec)
  gridy <- seq((min(Y)-perc*(max(Y)-min(Y))),(max(Y)+perc*(max(Y)-min(Y))),length=prec)
  
  fcomp <- array(0,c(prec,prec,k))
  for(h in 1:k){
    f <- function(gridx,gridy){  	  	
      r <- x$prior[h]*dnorm(gridx,mean=muX[h],sd=sdX[h])*dnorm(gridy,mean=polyfun(gridx,beta[h,]),sd=sdY[h])
    }
    fcomp[,,h] <- outer(gridx,gridy,f) 
  }
  fmisc <- matrix(0,prec,prec)
  for(h in 1:k) fmisc <- fmisc + fcomp[,,h]
  
  par(cex.axis = 0.3)   
  par(cex.lab = 1.2)
  persp(x = gridx, y = gridy, z = fmisc,
        main="", 
        xlab = "x",
        ylab = "y",
        zlab = "joint density", 
        theta = theta, 
        axes = TRUE, 
        box = TRUE
  ) 
      
}

.CW.plot <- function(x,nbin=35,contour=contour,nlevels=21,...)
{
  X       <- x$X
  Y       <- x$Y
  n       <- x$n
  k       <- x$k
  r       <- x$r
  group   <- x$group
  prior   <- x$prior
  beta    <- x$beta
  muX     <- x$muX
  sdX     <- x$sdX
  beta    <- x$beta
  sdY     <- x$sdY
  
  prec    <- 101
  perc    <- 0.1
  gridx   <- seq((min(X)-perc*(max(X)-min(X))),(max(X)+perc*(max(X)-min(X))),length=prec)
  gridy   <- seq((min(Y)-perc*(max(Y)-min(Y))),(max(Y)+perc*(max(Y)-min(Y))),length=prec)
  
  fcomp <- array(0,c(prec,prec,k))
  for(j in 1:k){
    f <- function(gridx,gridy){        
      r <- prior[j]*dnorm(gridx,mean=muX[j],sd=sdX[j])*dnorm(gridy,mean=polyfun(gridx,par=beta[j,]),sd=sdY[j])
    }
    fcomp[,,j] <- outer(gridx,gridy,f) 
  }
  fmisc <- matrix(0,prec,prec)
  for(h in 1:k) 
    fmisc <- fmisc + fcomp[,,h]
  
  par(mai=c(0.82,0.78,0.4,0.024)) 
  par(omi=c(0,0,0,0))
  
  # ----------------------------------
  # set up plotting area for histogram
  # ----------------------------------
  
  par(fig=c(0,1,0.65,1))
  par(cex.axis=0.6) 
  
  mixfun <- function(x,pesi=prior,m=muX,s=sdX){
    iniz <- 0
    for(j in 1:length(pesi))
      iniz <- pesi[j]*dnorm(x,m[j],s[j]) + iniz
    return(iniz)
  }
  
  plot(hist(X,breaks=seq(min(X),max(X),length=nbin)),freq=F,main="",xlab="") #,breaks=seq(min(X),max(X),length=35) ,xlab=expression(italic(x))
  curve(mixfun(x),add=T,lwd=2,n=1001)
  for(h in 1:k)
    curve(prior[h]*dnorm(x,mean=muX[h],sd=sdX[h]),add=T,lty=h,n=1001,col=h) 
  for(h in 1:k) 
    rug(X[which(group==h)],col=h)
  
  # ------------------------------------
  # set up plotting area for scatterplot
  # ------------------------------------
  
  par(fig=c(0,1,0,0.75),new=TRUE)
  
  plot(X,Y,col=group,cex=0.6,pch=group,lwd=2)
  if(contour==TRUE)
    contour(x = gridx, y = gridy, z = fmisc, nlevels = nlevels, col="lightgray", drawlabels=TRUE, add=T) 
  points(X, Y, cex=0.6, pch=group, col=group)
  for(h in 1:k){ 
    griglia <- seq(min(X[which(group==h)]),max(X[which(group==h)]),length=101)
    points(griglia,polyfun(griglia,par=beta[h,]),lty=h,col=h,type="l")
  }      
    
}

.scatter.plot <- function(x,contour=contour,nlevels=21,...)
{
  X       <- x$X
  Y       <- x$Y
  n       <- x$n
  k       <- x$k
  r       <- x$r
  group   <- x$group
  prior   <- x$prior
  beta    <- x$beta
  muX     <- x$muX
  sdX     <- x$sdX
  beta    <- x$beta
  sdY     <- x$sdY
  
  prec    <- 101
  perc    <- 0.1
  
  gridx <- seq((min(X)-perc*(max(X)-min(X))),(max(X)+perc*(max(X)-min(X))),length=prec)
  gridy <- seq((min(Y)-perc*(max(Y)-min(Y))),(max(Y)+perc*(max(Y)-min(Y))),length=prec)
  
  fcomp <- array(0,c(prec,prec,k))
  for(h in 1:k){
    f <- function(gridx,gridy){        
      r <- prior[h]*dnorm(gridx,mean=muX[h],sd=sdX[h])*dnorm(gridy,mean=polyfun(gridx,beta[h,]),sd=sdY[h])
    }
    fcomp[,,h] <- outer(gridx,gridy,f) 
  }
  fmisc <- matrix(0,prec,prec)
  for(h in 1:k) 
    fmisc <- fmisc + fcomp[,,h]
  
  par(mai=c(1.1,1.1,0.01,0.01))
  plot(X,Y,col=group,cex=0.6,pch=group,lwd=2)
  if(contour==TRUE)
    contour(x = gridx, y = gridy, z = fmisc, nlevels = nlevels, col="lightgray", drawlabels=TRUE, add=T) 
  points(X, Y, cex=0.6, pch=group, col=group)
  for(h in 1:k){ 
    griglia <- seq(min(X[which(group==h)]),max(X[which(group==h)]),length=101)
    points(griglia,polyfun(griglia,beta[h,]),lty=h,col=h,type="l")
  }      
  
}

